class Schemes:
    SSH = "ssh"
    HDFS = "hdfs"
    S3 = "s3"
    AZURE = "azure"
    HTTP = "http"
    HTTPS = "https"
    GS = "gs"
    GDRIVE = "gdrive"
    LOCAL = "local"
    OSS = "oss"
    WEBDAV = "webdav"
    WEBDAVS = "webdavs"
